package data;

import java.util.*;

/**
 * Purely functional balanced binary tree map.
 */
public class AAMap<Key extends Comparable<Key>,Value> implements Iterable<Value>, BinaryTree<Value> {
  private AAMap<Key,Value> left, right;
  private int level;

  private Key key;
  private Value value;

  public AAMap() {
    left = this;
    right = this;
  }
  public AAMap(Key k, Value v) {
    left = new AAMap<Key,Value>();
    right = new AAMap<Key,Value>();
    key = k;
    value = v;
    level = 1;
  }
  private AAMap(Key k, Value v, AAMap<Key,Value> l, AAMap<Key,Value> r, int lev) {
    key = k;
    value = v;
    left = l;
    right = r;
    level = lev;
  }

  public boolean isEmpty() {
    return level==0;
  }
  public AAMap<Key,Value> put(Key k, Value v) {
    if (level==0) return new AAMap<Key,Value>(k, v);
    AAMap<Key,Value> t = new AAMap<Key,Value>(key, value, left, right, level);
    int cmp = k.compareTo(key);
    if (cmp < 0) {
      t.left = t.left.put(k, v);
    } else if (cmp > 0) {
      t.right = t.right.put(k, v);
    } else {
      t.key = k;
      t.value = v;
      return t;
    }
    t = t.skew();
    t = t.split();
    return t;
  }
  public Value get(Key k) {
    if (level==0) return null;
    int cmp = k.compareTo(key);
    if (cmp < 0) return left.get(k);
    if (cmp > 0) return right.get(k);
    return value;
  }
  public AAMap<Key,Value> remove(Key k) {
    if (level==0) return this;
    AAMap<Key,Value> t = new AAMap<Key,Value>(key, value, left, right, level);
    int cmp = k.compareTo(key);
    if (cmp < 0) {
      t.left = t.left.remove(k);
    } else if (cmp > 0) {
      t.right = t.right.remove(k);
    } else {
      if (right.level==0) return new AAMap<Key,Value>();
      if (left.level==0) {
        AAMap<Key,Value> l = t.successor();
        t.right = t.right.remove(l.key);
        t.key = l.key;
        t.value = l.value;
      } else {
        AAMap<Key,Value> l = t.predecessor();
        t.left = t.left.remove(l.key);
        t.key = l.key;
        t.value = l.value;
      }
    }

    t.decreaseLevel();
    t = t.skew();
    t.right = t.right.skew();
    t.right.right = t.right.right.skew();
    t = t.split();
    t.right = t.right.split();
    return t;
  }

  private AAMap<Key,Value> skew() {
    AAMap<Key,Value> l = left;
    if (level!=0 && level == l.level) {
      left = l.right;
      l.right = this;
      return l;
    }
    return this;
  }
  private AAMap<Key,Value> split() {
    AAMap<Key,Value> r = right;
    if (level!=0 && level==r.right.level) {
      right = r.left;
      r.left = this;
      ++r.level;
      return r;
    }
    return this;
  }
  private AAMap<Key,Value> successor() {
    AAMap<Key,Value> t = right;
    while(t.left.level > 0) t = t.left;
    return t;
  }
  private AAMap<Key,Value> predecessor() {
    AAMap<Key,Value> t = left;
    while(t.right.level > 0) t = t.right;
    return t;
  }
  private void decreaseLevel() {
    int s = Math.min(left.level, right.level)+1;
    if (s < level) {
      level = s;
      if (s < right.level)
        right.level = s;
    }
  }

  public Iterator<Value> iterator() {
    if (level==0) return new BinaryTreeIterator<Value>();
    return new BinaryTreeIterator<Value>(this);
  }
  public Value getValue() {
    return value;
  }
  public AAMap<Key,Value> getLeft() {
    if (left.level==0) return null;
    return left;
  }
  public AAMap<Key,Value> getRight() {
    if (right.level==0) return null;
    return right;
  }

  public String toString() {
    if (level==0) return "[]";
    return "["+left+" , "+key+":"+value+" , "+right+"]";
  }
}
